/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/
#define FILECODE "F(302)"

extern "C" {
/*
 * task.c --
 *
 *      Task initialization and switching routines between the host
 *      and the monitor. 
 *     
 *      A task switch: 
 *          -saves the EFLAGS,CR0,CR2,CR4, and IDT
 *          -jumps to code on the shared page
 *              which saves the registers, GDT and CR3
 *              which then restores the registers, GDT and CR3
 *          -restores the IDT,CR0,CR2,CR4 and EFLAGS
 *                
 *      This file is pretty much independent of the host OS.
 *      
 */

/* On Linux, must come before any inclusion of asm/page.h --hpreg */
#include "hostKernel.h"
#ifdef linux
#   include <linux/string.h>
#   include <linux/kernel.h>

#   ifdef USE_PERFCTRS_HOSTED
#      include "perfctr.h"
#   endif

#   define EXPORT_SYMTAB
#endif

#include "vmware.h"
#include "modulecall.h"
#include "vmx86.h"
#include "task.h"
#include "vm_asm.h"
#include "cpuid.h"
#include "hostif.h"
}

typedef struct SysenterState {
   uint32       esp;
   uint32       eip;
   Selector     cs;
} SysenterState;


/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static Bool hostMayUseSysenter = FALSE;
template <class VMCP> static int 
Task_InitCrosspage_TOT(VMDriver *vm,          // IN
                       InitBlock *initParams) // IN: Initial params from the VM 
{
   Vcpuid vcpuid;

   ASSERT(sizeof (VMCrossPage) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
#ifdef MODULE_X86_64
      uint64 cr3reg;
#else
      uint32 cr3reg;
#endif
      LA linearAddr;

      VMCP *p = (VMCP*)HostIF_MapCrossPage(vm, (void *)(uintptr_t)initParams->crosspage[vcpuid], 
					   &linearAddr);
#ifdef MODULE_X86_64
      MPN crossPageMPN = HostIF_LookupUserMPN(vm, (void *)(uintptr_t)initParams->crosspage[vcpuid]);
      
      if ((int64)crossPageMPN <= 0) {
         return 1;
      }
#endif
      if (p == NULL) {
         return 1;
      }

      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
#ifdef MODULE_X86_64
      p->hostContext64.context.cr3 = cr3reg;
      p->hostVA64 = (uintptr_t) p;
      /*
       * In order to make the transition to legacy mode, we must first turn
       * paging off on the processor. For this we need to be executing
       * code on an identity mapped page (the crosspage).
       */
      p->crosspageLinearAddr = (LA)MPN_2_MA(crossPageMPN);

      /* Make sure there is no truncation */
      ASSERT((VA)p->crosspageLinearAddr == MPN_2_MA(crossPageMPN));      
      p->hostContext64VA = (uintptr_t)&p->hostContext64;
#else
      p->hostContext.task.cr3 = cr3reg;
      p->crosspageLinearAddr = linearAddr;
      p->hostContextVA = (uintptr_t)&p->hostContext;
#endif      

      {
         DTR idtrReg;
         GET_IDT(idtrReg);
#ifdef MODULE_X86_64
         /* XXX hack the initial monitor idt */
         p->monContext.idtr.dtr.offset = 0;
         p->monContext.idtr.dtr.limit = sizeof(Gate)*0x100 -1;
#else         
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
#endif
      }
#ifdef MODULE_X86_64
      {
         p->hostContext64.switchGdtr64.offset = p->crosspageLinearAddr + 
            offsetof(VMCrossPage, switchGdt);
         p->hostContext64.switchGdtr64.limit = sizeof(Descriptor) * 4 - 1;
         
         // null descriptor
         p->switchGdt[0].word[0] = 0;
         p->switchGdt[0].word[1] = 0;

         // 32 bit compatibility mode CS descriptor
         Desc_SetDescriptor(&p->switchGdt[1].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 1, 1);
         
         // 64 bit CS descriptor
         Desc_SetDescriptor(&p->switchGdt[2].desc, 0, (VA32)0xfffff, CODE_DESC, 
                            1, 0, 1, 0, 1);
         // Set the L bit to indicate running 64 bit mode.
         p->switchGdt[2].desc.longmode = 1;

         // 32 bit compatibility mode DS descriptor
         Desc_SetDescriptor(&p->switchGdt[3].desc, 0, (VA32)0xfffff, DATA_DESC, 
                            1, 0, 1, 1, 1);

         // Set up sel:offset for jumping to 32 bit code.
         // The code offset will be set up by the emit code in monitorHosted().
         p->hostContext64.jump32Code.selector = MAKE_SELECTOR(1, SELECTOR_GDT, 0);

         // Set up sel:offset for jumping to 64 bit code.
         // The code offset will be set up by the emit code in monitorHosted().
         p->hostContext64.jump64Code.selector = MAKE_SELECTOR(2, SELECTOR_GDT, 0);

         // Set up segment DS for running code in 32bit compability mode.
         p->hostContext64.switchDS = MAKE_SELECTOR(3, SELECTOR_GDT, 0);
      }
#endif
      
      {
         uint16 trReg;
         GET_TR(trReg);
#ifdef MODULE_X86_64
         p->hostContext64.tr = trReg;
#else
         p->hostContext.tr = trReg;
#endif
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;
      p->userCallRequest = MODULECALL_USERCALL_NONE;
      p->moduleCallInterrupted = FALSE;
      p->tscAdjustment = 0;
   }
   return 0;
}

/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage_V4
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

static int 
Task_InitCrosspage_V4(VMDriver *vm,          // IN
                      InitBlock *initParams) // IN: Initial params from the VM 
{
   LA baseLinearAddr = HOST_KERNEL_VA_2_LA(0);
   uint32 cr3reg;
   Vcpuid vcpuid;

   ASSERT(sizeof (VMCrossPage) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      /*
       * The physical memory associated to the crosspage has already been
       * locked in MonitorAllocateMemory() --hpreg
       */
      VMCrossPageV4 *p = (VMCrossPageV4*)HostIF_UserToDriverPtr(vm, initParams->crosspage[vcpuid]);
      if (p == NULL) {
         return 1;
      }
      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      p->hostContext.task.cr3 = cr3reg;
      p->hostVA = p;
      p->hostContextVA = (uintptr_t)&p->hostContext;
      p->crosspageLinearAddr = baseLinearAddr + (uintptr_t)p;
      {
         DTR idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
      }

      {
         uint16 trReg;
         GET_TR(trReg);
         p->hostContext.tr = trReg;
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;

      p->userCallRequest = MODULECALL_USERCALL_NONE;
      p->moduleCallInterrupted = FALSE;
    }
   return 0;
}

static Bool
DisableNMIDelivery(volatile uint32 *regPtr) // IN/OUT
{
   uint32 reg;
   
   reg = *regPtr;
   if ((APIC_LVT_DELVMODE(reg) == APIC_LVT_DELVMODE_NMI) &&
       (! APIC_LVT_ISMASKED(reg))) {
      *regPtr = reg | APIC_LVT_MASK;
      return TRUE;
   }
   return FALSE;
}

#define NMI_LINT0	(0x01)
#define NMI_LINT1	(0x02)
#define NMI_TIMER	(0x04)
#define NMI_ERR		(0x08)
#define NMI_PC		(0x10)
#define NMI_THERM	(0x20)

static unsigned int
DisableNMI2(VMDriver *vm)   // IN
{
   unsigned int nmi = 0;
   if (vm->hostAPIC) {
      switch (APIC_MAX_LVT(vm->hostAPIC)) {
         default:
         case 5:
	    if (DisableNMIDelivery(&APIC_THERM_REG(vm->hostAPIC))) {
	       nmi |= NMI_THERM;
	    }
	 case 4:
	    if (DisableNMIDelivery(&APIC_PC_REG(vm->hostAPIC))) {
	       nmi |= NMI_PC;
	    }
	 case 3:
	 case 2:
	 case 1:
	 case 0:
	    if (DisableNMIDelivery(&APIC_LINT0_REG(vm->hostAPIC))) {
	       nmi |= NMI_LINT0;
	    }
	    if (DisableNMIDelivery(&APIC_LINT1_REG(vm->hostAPIC))) {
	       nmi |= NMI_LINT1;
	    }
	    if (DisableNMIDelivery(&APIC_TIMER_REG(vm->hostAPIC))) {
	       nmi |= NMI_TIMER;
	    }
	    if (DisableNMIDelivery(&APIC_ERR_REG(vm->hostAPIC))) {
	       nmi |= NMI_ERR;
	    }
	    break;
      }
   }
   return nmi;
}

static unsigned int
DisableNMI(VMDriver *vm)   // IN
{
   unsigned int nmi = DisableNMI2(vm);
   unsigned int nmi2;
   
   /* Notify about race between timer and NMI on P4. It disables PC interrupt
      delivery when timer triggers, but actual delivery of NMI can race with us,
      and NMI handler can enable it under us. So we do it twice, hoping for
      the best... */
   nmi2 = DisableNMI2(vm);
   return nmi | nmi2;
}


static void
RestoreNMI(VMDriver *vm,  // IN
           unsigned int nmi) // IN
{
#define RestoreNMIDelivery(cond, apicr)				\
   do {								\
      if (nmi & cond) {						\
         uint32 reg;						\
	 							\
         reg = apicr;						\
         apicr = reg & ~APIC_LVT_MASK;				\
      }								\
   } while (0)

   RestoreNMIDelivery(NMI_LINT0, APIC_LINT0_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_LINT1, APIC_LINT1_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_TIMER, APIC_TIMER_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_ERR,   APIC_ERR_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_PC,    APIC_PC_REG(vm->hostAPIC));
   RestoreNMIDelivery(NMI_THERM, APIC_THERM_REG(vm->hostAPIC));
#undef RestoreNMIDelivery
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnToMonitor  --
 *
 *     This function saves the host OS' sysenter state prior to
 *     returning to the monitor.
 *
 *     The monitor does not know whether the host uses the MSRs. Thus,
 *     in order to avoid unnecessary writes to MSR_SYSENTER_CS, we
 *     clear this MSR here if the monitor does not use it but the host
 *     has set it to a non-zero value.
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnToMonitor(SysenterState *hostState,  // OUT
                        int monUsesSysenter)  // IN
{
   if (!hostMayUseSysenter) return;
   hostState->cs = (Selector)HostIF_RDMSR(MSR_SYSENTER_CS);
   if (hostState->cs != 0) {
      if (monUsesSysenter) {
         hostState->esp = (uint32)HostIF_RDMSR(MSR_SYSENTER_ESP);
         hostState->eip = (uint32)HostIF_RDMSR(MSR_SYSENTER_EIP);
      } else {
         /* Monitor does not use sysenter, but host does.  Clear
          * MSR_SYSENTER_CS to induce #GP if sysenter used in the
          * vmm/guest.
          */
         HostIF_WRMSR(MSR_SYSENTER_CS, 0);
      }
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SysenterReturnFromMonitor  --
 *
 *      Restore host sysenter state
 *
 *-----------------------------------------------------------------------------
 */

static INLINE_SINGLE_CALLER void
SysenterReturnFromMonitor(SysenterState *hostState,  // IN
                          int monUsesSysenter)       // IN
{
   if (!hostMayUseSysenter) return;
   if (monUsesSysenter || hostState->cs != 0) {
      HostIF_WRMSR(MSR_SYSENTER_CS, hostState->cs);
   }
   if (monUsesSysenter && hostState->cs != 0) {
      HostIF_WRMSR(MSR_SYSENTER_ESP, hostState->esp);
      HostIF_WRMSR(MSR_SYSENTER_EIP, hostState->eip);
   }
}


/*
 *-----------------------------------------------------------------------------
 *
 * SwitchToMonitor --
 *
 *      Wrapper that calls code to switch from the host to the monitor.
 *
 *      The wrapper ends with a long jump after having
 *      pushed EAX and EBX on the stack. Compensate for these
 *      two registers, and then for the retval and argument on
 *      on the stack
 *
 *      Note the dummy 4th argument on the stack compensates
 *      for the absence of segment information pushed on a
 *      near jump.
 *
 * Side effects:
 *      None for the module
 *
 * Bugs: EAX is caller-saved, no need to save/restore
 *       should use caller-saved EDX rather than EBX as temp
 *       in Task_Switch()
 *
 *-----------------------------------------------------------------------------
 */
#ifdef MODULE_X86_64

static INLINE_SINGLE_CALLER void
SwitchToMonitor(VMCrossPage *crosspage)
{
   PA monContextPA   = crosspage->crosspageLinearAddr + 
                       offsetof(VMCrossPage, monContext);
   VA32 monContextVA = (VPN_2_VA(MODULECALL_CROSS_PAGE_START)
                       + offsetof(VMCrossPage,monContext));
   __asm__ __volatile__(
      "call *%4"       "\n\t"
      :
      : "a" (&crosspage->hostContext64),
        "b" (monContextPA),
        "c" (monContextVA),
        "d" (crosspage->hostSwitchCR3),
        "g" (crosspage->contextSwitchCode)
      : "memory", "cc"
   );
}

#else

static INLINE_SINGLE_CALLER void
SwitchToMonitor(unsigned char *codePtr,
                ContextInfo *src,
                ContextInfo *dst,
                uint32 newDstVA)
{
#ifdef __GNUC__
   uint32 dummy1;
   uint32 dummy2;

   /* Checked against the Intel manual and GCC --hpreg */
   __asm__ __volatile__(
      "pushl %2"       "\n\t"
      "pushl %3"       "\n\t"
      "pushl %4"       "\n\t"
      "pushl %5"       "\n\t"
      "call *%6"       "\n\t"
      "popl %0"        "\n\t"
      "popl %1"        "\n\t"
      "addl %7, %%esp"
      : "=b" (dummy1),
        "=a" (dummy2)
      : "g" (newDstVA),
        "ri" (dst),
        "ri" (src),
        "ri" (0),
        "r" (codePtr),
        "i" ((1 + 1 + 3) * 4) // retval + segment + 3 arguments pushed
      : "memory", "cc"
   );
#elif _MSC_VER
   __asm push newDstVA
   __asm push dst
   __asm push src
   __asm push 0
   __asm call codePtr
   __asm pop ebx
   __asm pop eax
   __asm add esp, 20
#else
#   error No compiler defined for SwitchToMonitor
#endif
}

#endif // MODULE_X86_64

/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch_TOT(VMDriver *vm,  // IN
                Vcpuid vcpuid) // IN
{
#ifdef MODULE_X86_64
   uint64      flags, cr0reg, cr2reg, cr4reg;
   Selector    cs;
   uint64      gs64, fs64;
   uint64      drReg;
#else
   DTR         gdtreg;
   uint32      flags,cr0reg, cr2reg,cr4reg;
   Selector    cs, gs, fs;
   uint32      drReg;
#endif
   Descriptor *desc;
   SysenterState sysenterState;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
   VMCrossPageTOT *crosspage = (VMCrossPageTOT*)vm->crosspage[vcpuid];
   uint32 new_cr4;

   /*
    * monContextVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
   uint32 monContextVA = (uint32)
      (VPN_2_VA(MODULECALL_CROSS_PAGE_START_TOT)+ offsetof(VMCrossPageTOT,monContext)); 
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   vm->currentHostCpu[vcpuid] = HostIF_GetCurrentPCPU();
   
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);
   
   /* Ensure global pages are flushed */
   new_cr4 = cr4reg & ~CR4_PGE;
   SET_CR4(new_cr4);

   crosspage->switchHostCR4 = new_cr4;
   
   /*
    * Save DR7 since we need to disable debug breakpoints during the
    * world switch code.  Save DR6 in order to accomodate the ICEBP
    * instruction.  All other debug registers are saved lazily by the
    * monitor prior to first use.  NOTE: Since monitor is in legacy
    * mode save all DRs for x86_64 here.
    */

   crosspage->hostDRSaved = 0;

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drReg); \
                crosspage->hostDR[n] = drReg; \
                crosspage->shadowDR[n] = (uint32) drReg; \
                crosspage->hostDRSaved |= (1 << n);

#ifdef MODULE_X86_64
   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
#endif // MODULE_X86_64

   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);
   
#undef SAVE_DR_SHADOW
   
   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(crosspage->hostDR[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

#ifdef MODULE_X86_64
   GET_GDT64(crosspage->hostContext64.gdtr64.dtr);
#else
   GET_GDT(gdtreg);
#endif
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
#ifdef MODULE_X86_64
   gs64 = GET_GS64();
   fs64 = GET_FS64();
#else
   gs = GET_GS();
   fs = GET_FS();
#endif
   
   SysenterReturnToMonitor(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
#ifdef MODULE_X86_64
   crosspage->hostContext64.tr = trReg;
#else
   crosspage->hostContext.tr = trReg;
#endif
   if (trReg) {
#ifdef MODULE_X86_64
      vAddr = HOST_KERNEL_LA_2_VA(crosspage->hostContext64.gdtr64.dtr.offset) + trReg;
#else      
      vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
#endif     
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */
   cs = GET_CS();
#ifdef MODULE_X86_64
   crosspage->hostContext64.context.cs = cs;
#else
   *(uint16*) (&crosspage->hostContext.jump[4]) = cs;
   crosspage->hostContext.task.cs = cs;
#endif   
   /*
    * See comment for SwitchToMonitor()
    */
   SwitchToMonitor(&crosspage->contextSwitchCode[0], &crosspage->hostContext, &crosspage->monContext, monContextVA);

   SysenterReturnFromMonitor(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);
   
#ifdef MODULE_X86_64
   SET_FS64(fs64);
   SET_GS64(gs64);
#else
   SET_FS(fs);
   SET_GS(gs);
#endif

#ifdef MODULE_X86_X64
#define CAST_HOST_DR(x) (x)
#else
#define CAST_HOST_DR(x) ((uint32) x)
#endif

#define RESTORE_DR(n) if ((crosspage->hostDRSaved & (1 << n)) && \
                          (crosspage->hostDR[n] != crosspage->shadowDR[n])) { \
                          drReg = CAST_HOST_DR(crosspage->hostDR[n]); \
                          SET_DR##n(drReg); \
                      }

   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR
#undef CAST_HOST_DR
   
   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_TOT_INTR) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   vm->currentHostCpu[vcpuid] = INVALID_HOST_CPU;

   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch_V45(VMDriver *vm,  // IN
                Vcpuid vcpuid) // IN
{
#ifdef MODULE_X86_64
   uint64      flags, cr0reg, cr2reg, cr4reg;
   Selector    cs;
   uint64      gs64, fs64;
   uint64      drReg;
#else
   DTR         gdtreg;
   uint32      flags,cr0reg, cr2reg,cr4reg;
   Selector    cs, gs, fs;
   uint32      drReg;
#endif
   Descriptor *desc;
   SysenterState sysenterState;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
   VMCrossPageV45 *crosspage = (VMCrossPageV45*)vm->crosspage[vcpuid];
   uint32 new_cr4;

   /*
    * monContextVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
   uint32 monContextVA = (uint32)
      (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V45)+ offsetof(VMCrossPageV45,monContext)); 
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   vm->currentHostCpu[vcpuid] = HostIF_GetCurrentPCPU();
   
   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);
   
   /* Ensure global pages are flushed */
   new_cr4 = cr4reg & ~CR4_PGE;
   SET_CR4(new_cr4);

   crosspage->switchHostCR4 = new_cr4;
   
   /*
    * Save DR7 since we need to disable debug breakpoints during the
    * world switch code.  Save DR6 in order to accomodate the ICEBP
    * instruction.  All other debug registers are saved lazily by the
    * monitor prior to first use.  NOTE: Since monitor is in legacy
    * mode save all DRs for x86_64 here.
    */

   crosspage->hostDRSaved = 0;

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(drReg); \
                crosspage->hostDR[n] = drReg; \
                crosspage->shadowDR[n] = (uint32) drReg; \
                crosspage->hostDRSaved |= (1 << n);

#ifdef MODULE_X86_64
   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
#endif // MODULE_X86_64

   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);
   
#undef SAVE_DR_SHADOW
   
   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(crosspage->hostDR[7] & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

#ifdef MODULE_X86_64
   GET_GDT64(crosspage->hostContext64.gdtr64.dtr);
#else
   GET_GDT(gdtreg);
#endif
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
#ifdef MODULE_X86_64
   gs64 = GET_GS64();
   fs64 = GET_FS64();
#else
   gs = GET_GS();
   fs = GET_FS();
#endif
   
   SysenterReturnToMonitor(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
#ifdef MODULE_X86_64
   crosspage->hostContext64.tr = trReg;
#else
   crosspage->hostContext.tr = trReg;
#endif
   if (trReg) {
#ifdef MODULE_X86_64
      vAddr = HOST_KERNEL_LA_2_VA(crosspage->hostContext64.gdtr64.dtr.offset) + trReg;
#else      
      vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
#endif     
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */
   cs = GET_CS();
#ifdef MODULE_X86_64
   crosspage->hostContext64.context.cs = cs;
#else
   *(uint16*) (&crosspage->hostContext.jump[4]) = cs;
   crosspage->hostContext.task.cs = cs;
#endif   
   /*
    * See comment for SwitchToMonitor()
    */
   SwitchToMonitor(&crosspage->contextSwitchCode[0], &crosspage->hostContext, &crosspage->monContext, monContextVA);

   SysenterReturnFromMonitor(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);
   
#ifdef MODULE_X86_64
   SET_FS64(fs64);
   SET_GS64(gs64);
#else
   SET_FS(fs);
   SET_GS(gs);
#endif

#ifdef MODULE_X86_X64
#define CAST_HOST_DR(x) (x)
#else
#define CAST_HOST_DR(x) ((uint32) x)
#endif

#define RESTORE_DR(n) if ((crosspage->hostDRSaved & (1 << n)) && \
                          (crosspage->hostDR[n] != crosspage->shadowDR[n])) { \
                          drReg = CAST_HOST_DR(crosspage->hostDR[n]); \
                          SET_DR##n(drReg); \
                      }

   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR
#undef CAST_HOST_DR
   
   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_V45_INTR) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   vm->currentHostCpu[vcpuid] = INVALID_HOST_CPU;

   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}


/*
 *----------------------------------------------------------------------
 * Task_ProbeSysenterMSR --
 *
 *    This function computes sysenterMSR. Note that we assume that if a
 *    host is ever going to use the sysenter MSRs, they will have been
 *    initialized here; if some future host lazily sets up the sysenter
 *    MSRs, this code will cause trouble.
 *
 *----------------------------------------------------------------------
 */
void
Task_ProbeSysenterMSR(void)
{
   CPUOem cpuOem = CPUID_GetOEM();
   uint32 features = CPUID_GetFeatures();
   uint32 version = CPUID_GetVersion();

   /*
    * Some old intel CPUs set the SEP bit, even though they are not
    * capable of sysenter. In particular, CPUs with family of 6, but
    * with the whole version string < 0x633 (family, model, stepping).
    * This check is also performed in vmx/vmcore/cpuid.c.
    */
   if (cpuOem == CPUID_OEM_INTEL && CPUID_FAMILY(version) == 6 &&
       (CPUID_MODEL(version) < 3 ||
        (CPUID_MODEL(version) == 3 && CPUID_STEPPING(version) < 3))) {

      Warning("VMMON: CPU too old to use sysenter (ok).\n");
   } else if (features & CPUID_FEATURE_COMMON_SEP) {
      /* The CPU appears to support sysenter. */
      hostMayUseSysenter = 1;
   }
}

/*
 *------------------------------------------------------------------------------
 *
 * Task_InitCrosspage
 *
 *    Initialize the crosspage used to switch to the monitor task. 
 *
 * Results:
 *    0 on success
 *    != 0 on failure
 *
 * Side effects:
 *    None
 *  
 *------------------------------------------------------------------------------
 */

template <class VMCP> static int 
Task_InitCrosspage(VMDriver *vm,          // IN
                   InitBlock *initParams) // IN: Initial params from the VM 
{
   LA baseLinearAddr = HOST_KERNEL_VA_2_LA(0);
   uint32 cr3reg;
   Vcpuid vcpuid;

   ASSERT(sizeof(VMCP) < PAGE_SIZE);
   ASSERT(MODULECALL_CROSS_PAGE_LEN==1);
   for (vcpuid = 0; vcpuid < initParams->numVCPUs;  vcpuid++) {
      /*
       * The physical memory associated to the crosspage has already been locked
       * in MonitorAllocateMemory() --hpreg
       */
      VMCP *p = (VMCP*)HostIF_UserToDriverPtr(vm, initParams->crosspage[vcpuid]);
      if (p == NULL) {
         return 1;
      }
      vm->crosspage[vcpuid] = p;
      GET_CR3(cr3reg);
      p->hostContext.task.cr3 = cr3reg;
      p->hostVA = p;
      p->hostContextVA = (uintptr_t)&p->hostContext;
      p->crosspageLinearAddr = baseLinearAddr + (uintptr_t)p;
      {
         DTR idtrReg;
         GET_IDT(idtrReg);
         p->monContext.idtr.dtr = idtrReg; /* irrelevant */
      }

      {
         uint16 trReg;
         GET_TR(trReg);
         p->hostContext.tr = trReg;
      }

      p->irqRelocateOffset[0] = IRQ_HOST_INTR1_BASE; 
      p->irqRelocateOffset[1] = IRQ_HOST_INTR2_BASE;

#if defined(WINNT_DDK) && defined(VMX86_DEBUG)
      p->vmPtr = vm;
#endif
   }
   return 0;
}

int 
Task_InitCrosspage(VMDriver *vm,
                   InitBlock *initParams) // Initial params from the VM
{
   switch (vm->vmVersion) {
      case VME_TOT:
         return Task_InitCrosspage_TOT<VMCrossPageTOT>(vm, initParams);
      case VME_V45:
         return Task_InitCrosspage_TOT<VMCrossPageV45>(vm, initParams);
      case VME_V4:
         return Task_InitCrosspage_V4(vm, initParams);
      case VME_GSX251:
      case VME_GSX25:
         return Task_InitCrosspage<VMCrossPageGSX25>(vm, initParams);
      case VME_GSX2:
         return Task_InitCrosspage<VMCrossPageGSX2>(vm, initParams);
      case VME_V321:
         return Task_InitCrosspage<VMCrossPageV321>(vm, initParams);
      case VME_V32:
         return Task_InitCrosspage<VMCrossPageV32>(vm, initParams);
      case VME_V2:
      case VME_V3:
      case VME_GSX1:
         return Task_InitCrosspage<VMCrossPageV3>(vm, initParams);
      default:
         return 0xDEAD;
   }
}

/*
 *-----------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state.
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned.
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt.
 *
 *-----------------------------------------------------------------------------
 */

void 
Task_Switch_V4(VMDriver *vm,  // IN
               Vcpuid vcpuid) // IN
{
   uint32      monContextVA;     
   DTR         gdtreg;
   Descriptor *desc;
   uint32      flags,cr0reg, cr2reg,cr4reg;
   uint32      dr0reg, dr1reg, dr2reg, dr3reg, dr6reg, dr7reg;
   SysenterState sysenterState;
   Selector    gs, fs;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
   VMCrossPageV4 *crosspage = (VMCrossPageV4*)vm->crosspage[vcpuid];
  
   /*
    * monContextVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
   monContextVA = (uint32)
      (VPN_2_VA(MODULECALL_CROSS_PAGE_START_V4)+ offsetof(VMCrossPageV4,monContext)); 
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);

   /* Ensure global pages are flushed */
   SET_CR4(cr4reg & ~CR4_PGE);

   /*
    * We don't know what the host has been doing with the
    * debug registers, so we save them here and
    * conditionally restore them when we come back.
    *
    * We could do better by saving and restoring the
    * registers on the monitor side only when the monitor
    * needs to use them, but that will be a different
    * change.
    */

#define SAVE_DR_SHADOW(n) \
                GET_DR##n(dr##n##reg); \
                crosspage->shadowDR[n] = dr##n##reg;

   SAVE_DR_SHADOW(0);
   SAVE_DR_SHADOW(1);
   SAVE_DR_SHADOW(2);
   SAVE_DR_SHADOW(3);
   SAVE_DR_SHADOW(6);
   SAVE_DR_SHADOW(7);

#undef SAVE_DR_SHADOW

   /*
    * Disable debug exceptions during the switch.  Note: If
    * the DR7 GD bit is on, a trap occurred at the GET_DR0
    * above, and this process is most likely dead.
    */

   if (UNLIKELY(dr7reg & DR7_ENABLED)) {
      SET_DR7(DR7_DEFAULT);
      crosspage->shadowDR[7] = DR7_DEFAULT;
   }

   GET_GDT(gdtreg); 
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
   gs = GET_GS();
   fs = GET_FS();

   SysenterReturnToMonitor(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    */
   
   GET_TR(trReg);
   crosspage->hostContext.tr = trReg;
   if (trReg) {
     vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */

   *(uint16*) (&crosspage->hostContext.jump[4]) = GET_CS();
   crosspage->hostContext.task.cs = GET_CS();
   
   /*
    * See comment for SwitchToMonitor()
    */

   SwitchToMonitor(&crosspage->contextSwitchCode[0],
                  &crosspage->hostContext,
                  &crosspage->monContext,
                  monContextVA);

   SysenterReturnFromMonitor(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_LDT(ldtreg);  
   
   SET_FS(fs);
   SET_GS(gs);

#define RESTORE_DR(n) if (dr##n##reg != crosspage->shadowDR[n]) \
                              SET_DR##n(dr##n##reg)
   RESTORE_DR(0);
   RESTORE_DR(1);
   RESTORE_DR(2);
   RESTORE_DR(3);
   RESTORE_DR(6);
   RESTORE_DR(7);

#undef RESTORE_DR

   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_V4_INTR) {
      /*
       * Note we must do the RAISE_INTERRUPT before ever enabling
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw).
       * Note2: RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement.
       */

#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)

      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCTRS_HOSTED)
   PerfCtr_SnapshotFromMonitor();
#endif
}

/*
 *------------------------------------------------------------------------------
 *
 * Task_Switch --
 *
 *      Switches from the host context into the monitor
 *      context. Think of it as a coroutine switch that changes
 *      not only the registers, but also the address space
 *      and all the hardware state. 
 *
 * Results:
 *      Next module call (or user call for that matter) is
 *      returned. 
 *
 * Side effects:
 *      Jump to the other side. Has no direct effect on the
 *      host-visible state except that it might generate an interrupt. 
 *
 *------------------------------------------------------------------------------
 */

template <class VMCP> static void
Task_Switch(VMCP *crosspage,  // IN
            VMDriver *vm)     // IN
{
   DTR         gdtreg;
   Descriptor *desc;
   uint32      flags,cr0reg, cr2reg,cr4reg;
   uint32      dr0reg, dr1reg, dr2reg, dr3reg, dr6reg, dr7reg;
   SysenterState sysenterState;
   Selector    gs, fs;
   VA          vAddr;
   uint16 trReg;   
   uint16 ldtreg;
   unsigned int lint;
  
   /*
    * monContextVA points to the same location as
    * vm->crosspage->monContext, only refers to it
    * in the monitor's virtual address space
    */
#if defined(linux) && defined(USE_PERFCOUNTERS)
   PerfCtr_SnapshotFromHost();
#endif
   lint = DisableNMI(vm);
   SAVE_FLAGS(flags); 
   CLEAR_INTERRUPTS();

   /*
    * Save CR state (not CR3!)
    */
   GET_CR0(cr0reg);
   GET_CR2(cr2reg);
   GET_CR4(cr4reg);

   /* Ensure global pages are flushed */
   SET_CR4(cr4reg & ~CR4_PGE);

   GET_DR0(dr0reg);
   GET_DR1(dr1reg);
   GET_DR2(dr2reg);
   GET_DR3(dr3reg);
   GET_DR6(dr6reg);
   GET_DR7(dr7reg);

   GET_GDT(gdtreg); 
   GET_LDT(ldtreg); 

   /*
    * GS and FS are saved outside of the SwitchToMonitor() code to 1)
    * minimize the amount of code handled there and 2) prevent us from
    * faulting if they happen to be in the LDT (since the LDT is saved and
    * restored here too) and 3) make sure that if we do fault (because the
    * uncached descriptor for GS or FS's Selector has become invalid) we
    * fault in the host's context rather than the monitor or, worse,
    * never-never land. --Jeremy.
    */
   gs = GET_GS();
   fs = GET_FS();

   SysenterReturnToMonitor(&sysenterState, crosspage->monUsesSysenter);

   /*
    * Mark our task descriptor as unused, or we won't be
    * able to come back to it (SET_TR set the busy bit)
    * 
    */
   
   GET_TR(trReg);
   crosspage->hostContext.tr = trReg;
   if (trReg) {
     vAddr = HOST_KERNEL_LA_2_VA(gdtreg.offset) + trReg;
     desc = (Descriptor *)(vAddr);
     if (Desc_Type(desc) == TASK_DESC_BUSY) {
       Desc_SetType(desc, TASK_DESC);
     }
   }


   /*
    * Save the CS on this side
    */

   *(uint16*) (&crosspage->hostContext.jump[4]) = GET_CS();
   crosspage->hostContext.task.cs = GET_CS();
   
   /*
    * see comment in taskswitch.h 
    */

   SwitchToMonitor(&crosspage->contextSwitchCode[0],
                  &crosspage->hostContext,
                  &crosspage->monContext,
                  vm->monContextVA);

   SysenterReturnFromMonitor(&sysenterState, crosspage->monUsesSysenter);
   SET_CR0(cr0reg);
   SET_CR2(cr2reg);
   SET_CR4(cr4reg);
   
   SET_DR0(dr0reg);
   SET_DR1(dr1reg);
   SET_DR2(dr2reg);
   SET_DR3(dr3reg);
   SET_DR6(dr6reg);
   SET_DR7(dr7reg);

   SET_LDT(ldtreg);  
   
   SET_FS(fs);
   SET_GS(gs);
   
   ASSERT_NO_INTERRUPTS();

   if (crosspage->moduleCallType == MODULECALL_V3_INTR) {
      /*
       * Note we must do thje RAISE_INTERRUPT before ever enabling 
       * interrupts or bad things have happened (might want to know exactly
       * what bad things btw). 
       * Note2; RAISE_INTERRUPT() only takes an constant and hence with switch
       * statement. 
       */
#define IRQ_INT(_x) case _x: RAISE_INTERRUPT(_x); break 
#define IRQ_INT2(_x) IRQ_INT(_x); IRQ_INT(_x + 1)
#define IRQ_INT4(_x) IRQ_INT2(_x); IRQ_INT2(_x + 2)
#define IRQ_INT8(_x) IRQ_INT4(_x); IRQ_INT4(_x + 4)
#define IRQ_INT16(_x) IRQ_INT8(_x); IRQ_INT8(_x + 8)
#define IRQ_INT32(_x) IRQ_INT16(_x); IRQ_INT16(_x + 16)
      switch (crosspage->args[0]) {
	 // These are the general IO interrupts
	 // It would be nice to generate this dynamically, but see Note2 above.
	 IRQ_INT32(0x20);
	 IRQ_INT32(0x40);
	 IRQ_INT32(0x60);
	 IRQ_INT32(0x80);
	 IRQ_INT32(0xa0);
	 IRQ_INT32(0xc0);
	 IRQ_INT32(0xe0);

      default: 
	 /*
	  * XXXX nt
	  * running on a 2 processor machine we hit this Panic with int 0xD1 0x61 ...
	  */
	 Warning("Received Unexpected Interrupt: 0x%X in Task_Switch()\n", crosspage->args[0]);
	 Panic("Received Unexpected Interrupt: 0x%X\n", crosspage->args[0]);
      }
   }
   
   RESTORE_FLAGS(flags);
   RestoreNMI(vm, lint);
#if defined(linux) && defined(USE_PERFCOUNTERS)
   PerfCtr_SnapshotFromMonitor();
#endif
      
   return;
}

/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM  --
 *
 *      Let's move RunVM here from Vmx86... linux/signal.h contains
 *      code which does 'unsigned long int qq = -1', and this causes
 *      warnings... And we can make template products static then...
 *
 *----------------------------------------------------------------------
 */
 
/*
 *----------------------------------------------------------------------
 *
 * Vmx86_RunVM  --
 *
 *      Return to the monitor.
 *      Main interaction between the module and the monitor.
 *      Leaves the loop only to go back to user mode.
 *      Every interation of the loop switches back to the monitor
 *
 * Results:
 *      Returns the MODULECALL that forced it to leave the loop.
 *      back to the IOCTL handler of the module device driver
 *
 * Side effects:
 *      Dispatches the messages, everything can change
 *
 *----------------------------------------------------------------------
 */
template <class VMCP> static int
Vmx86_RunVM(VMCP *crosspage, VMDriver *vm)
{
   ModuleCallTypeV3 op;
   uint32 retval;

   ASSERT(vm && crosspage);

   retval = MODULECALL_V3_USERRETURN;

   while(1) {

      crosspage->retval = retval;

      /*
       * Task_Switch changes the world to the monitor.
       * The monitor is waiting in the BackToHost routine
       */
      Task_Switch(crosspage, vm);

      op = crosspage->moduleCallType;

      retval = MODULECALL_USER_START;

      if (op >  MODULECALL_USER_START&& op < MODULECALL_USER_END) {
         return op;
      }


      switch (op) {
      case MODULECALL_V3_NONE :
         break;


      case MODULECALL_V3_INTR : {
	 /*
	  * Note that as of SMP support, irq is the actual Interrupt number
	  * rather than the IRQ, b/c the IO APIC redirects IRQs every which
	  * way. --Jeremy.
	  */
#ifdef notdef
	 if (crosspage->args[1] == USERCALL_NONE) {
	    /* nothing to be done if MODULECALL_V3_USERRETURN */
	    retval = MODULECALL_V3_USERRETURN;
	    break;
	 }
#endif
	 return crosspage->args[1];
      }
      case MODULECALL_V3_ISMPNLOCKED : {
         MPN mpn = crosspage->args[0];
         retval = HostIF_IsMPNLocked(vm, mpn);
         break;
      }

      case MODULECALL_V3_LOCKPAGE: {
         char dummy;
         void *addr = (void *)crosspage->args[0];
	 ASSERT(sizeof retval == sizeof (MPN));
         retval = Vmx86_LockPage(vm, addr, FALSE);
	 if (PAGE_LOCK_SUCCESS(retval)) {
            if (HostIF_CopyFromUser(&dummy, addr, 1) == 0) {
               HostIF_CopyToUser(addr, &dummy, 1);
            } else {
               retval = Vmx86_UnlockPage(vm, addr);
	       if (PAGE_LOCK_SUCCESS(retval)) {
		  retval = PAGE_LOCK_TOUCH_FAILED;
	       }
            }
         }
         break;
      }

      case MODULECALL_V3_UNLOCKPAGE: {
         char dummy;
         void *addr = (void *)crosspage->args[0];
         Bool dirty = (Bool)crosspage->args[1];

         if (dirty) {
            if (HostIF_CopyFromUser(&dummy, addr, 1) != 0 ||
                HostIF_CopyToUser(addr, &dummy, 1) != 0) {
	       retval = PAGE_UNLOCK_TOUCH_FAILED;
            } else {
	       retval = Vmx86_UnlockPage(vm, addr);
	    }
         } else {
            retval = Vmx86_UnlockPage(vm, addr);
         }
         break;
      }

      default:
         Panic("ModuleCallLoop %d not supported \n",op);
      }
   }

   ASSERT(op >= MODULECALL_USER_START && op < MODULECALL_USER_END);
   return op;
}

int Vmx86_RunVM(VMDriver *vm, Vcpuid vcpuid) {
   void* crosspage = vm->crosspage[vcpuid];
   switch (vm->vmVersion) {
      case VME_V2:
      case VME_GSX1:
      case VME_V3:
         return Vmx86_RunVM((VMCrossPageV3*)crosspage, vm);
      case VME_V321:
      	 return Vmx86_RunVM((VMCrossPageV321*)crosspage, vm);
      case VME_V32:
         return Vmx86_RunVM((VMCrossPageV32*)crosspage, vm);
      case VME_GSX2:
         return Vmx86_RunVM((VMCrossPageGSX2*)crosspage, vm);
      case VME_GSX251:
      case VME_GSX25:
         return Vmx86_RunVM((VMCrossPageGSX25*)crosspage, vm);
      case VME_V4:
         return Vmx86_RunVM_V4(vm, vcpuid);
      case VME_V45:
         return Vmx86_RunVM_V45(vm, vcpuid);
      case VME_TOT:
         return Vmx86_RunVM_TOT(vm, vcpuid);
      default:
         return -EINVAL;
   }
}

/*
 *----------------------------------------------------------------------
 *
 * CheckPadding --
 *
 *      check for expected padding --
 *      this check currently fails on the egcs compiler
 *
 * Results:
 *
 *      TRUE if the check succeeds -- module will be loaded
 *
 *
 *
 * Side effects:
 *      output to kernel log on error
 *
 *----------------------------------------------------------------------
 */
int
Task_CheckPadding(void)
{
   DTRWords dtr;
   uint16 *x;

   memset(&dtr, 0, sizeof dtr);
   dtr.dtr.limit = 0x1111;
   dtr.dtr.offset = 0x22223333;

   x = (uint16*)&dtr;

   if (x[0] == 0x1111 && x[1] == 0x3333 && x[2] == 0x2222) {
   } else {
      Warning("DTR padding\n");
      goto error;
   }

   if (   sizeof (VMCrossPageTOT) != MODULECALL_CROSSPAGE_SIZE_TOT
       || offsetof(VMCrossPageTOT, crosspageLinearAddr)
             != MODULECALL_CROSSPAGE_LAOFFSET_TOT) {
      Warning("T: cross page 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof (VMCrossPageTOT), MODULECALL_CROSSPAGE_SIZE_TOT,
              offsetof(VMCrossPageTOT, crosspageLinearAddr),
              MODULECALL_CROSSPAGE_LAOFFSET_TOT);
      goto error;
   }

   if (   sizeof (VMCrossPageV45) != MODULECALL_CROSSPAGE_SIZE_V45
       || offsetof(VMCrossPageV45, crosspageLinearAddr)
             != MODULECALL_CROSSPAGE_LAOFFSET_V45) {
      Warning("T: cross page 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof (VMCrossPageV45), MODULECALL_CROSSPAGE_SIZE_V45,
              offsetof(VMCrossPageV45, crosspageLinearAddr),
              MODULECALL_CROSSPAGE_LAOFFSET_V45);
      goto error;
   }

   if (   sizeof (VMCrossPageV4) != MODULECALL_CROSSPAGE_SIZE_V4
       || offsetof(VMCrossPageV4, crosspageLinearAddr)
             != MODULECALL_CROSSPAGE_LAOFFSET_V4) {
      Warning("T: cross page 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof (VMCrossPageV4), MODULECALL_CROSSPAGE_SIZE_V4,
              offsetof(VMCrossPageV4, crosspageLinearAddr),
              MODULECALL_CROSSPAGE_LAOFFSET_V4);
      goto error;
   }

   if (sizeof(VMCrossPageGSX25) != MODULECALL_CROSSPAGE_SIZE_GSX25 ||
       offsetof(VMCrossPageGSX25,crosspageLinearAddr)!=MODULECALL_CROSSPAGE_LAOFFSET_GSX25) {
      Warning("T: cross page [gsx25] 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof(VMCrossPageGSX25), MODULECALL_CROSSPAGE_SIZE_GSX25,
              offsetof(VMCrossPageGSX25,crosspageLinearAddr), MODULECALL_CROSSPAGE_LAOFFSET_GSX25);
      goto error;
   }

   if (sizeof(VMCrossPageGSX2) != MODULECALL_CROSSPAGE_SIZE_GSX2 ||
       offsetof(VMCrossPageGSX2,crosspageLinearAddr)!=MODULECALL_CROSSPAGE_LAOFFSET_GSX2) {
      Warning("T: cross page [gsx2] 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof(VMCrossPageGSX2), MODULECALL_CROSSPAGE_SIZE_GSX2,
              offsetof(VMCrossPageGSX2,crosspageLinearAddr), MODULECALL_CROSSPAGE_LAOFFSET_GSX2);
      goto error;
   }

   if (sizeof(VMCrossPageV321) != MODULECALL_CROSSPAGE_SIZE_V321 ||
       offsetof(VMCrossPageV321,crosspageLinearAddr)!=MODULECALL_CROSSPAGE_LAOFFSET_V321) {
      Warning("T: cross page [v321] 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof(VMCrossPageV321), MODULECALL_CROSSPAGE_SIZE_V321,
              offsetof(VMCrossPageV321,crosspageLinearAddr), MODULECALL_CROSSPAGE_LAOFFSET_V321);
      goto error;
   }

   if (sizeof(VMCrossPageV32) != MODULECALL_CROSSPAGE_SIZE_V32 ||
       offsetof(VMCrossPageV32,crosspageLinearAddr)!=MODULECALL_CROSSPAGE_LAOFFSET_V32) {
      Warning("T: cross page [v32] 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof(VMCrossPageV32), MODULECALL_CROSSPAGE_SIZE_V32,
              offsetof(VMCrossPageV32,crosspageLinearAddr), MODULECALL_CROSSPAGE_LAOFFSET_V32);
      goto error;
   }

   if (sizeof(VMCrossPageV3) != MODULECALL_CROSSPAGE_SIZE_V3 ||
       offsetof(VMCrossPageV3,crosspageLinearAddr)!=MODULECALL_CROSSPAGE_LAOFFSET_V3) {
      Warning("T: cross page [v3] 0x%x expected 0x%x. la 0x%x expected 0x%x\n",
              sizeof(VMCrossPageV3), MODULECALL_CROSSPAGE_SIZE_V3,
              offsetof(VMCrossPageV3,crosspageLinearAddr), MODULECALL_CROSSPAGE_LAOFFSET_V3);
      goto error;
   }

   return TRUE;

error:
   printk("/dev/vmmon: Cannot load module. Use standard gcc compiler\n");
   return FALSE;
}
