/* **********************************************************
 * Copyright 1998 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/


/*
 * modulecall.h
 *
 *        Monitor <-->  Module (kernel driver) interface
 */

#ifndef _MODULECALL_H
#define _MODULECALL_H

#define INCLUDE_ALLOW_USERLEVEL
#define INCLUDE_ALLOW_VMM
#define INCLUDE_ALLOW_MODULE
#define INCLUDE_ALLOW_VMKERNEL
#include "includeCheck.h"

#include "x86.h"
#include "vm_time.h"
#include "vcpuid.h"
#include "vcpuset.h"

/*
 *----------------------------------------------------------------------
 *
 * ModuleCallType --
 *
 *      Enumaration of support calls done by the module
 *
 *----------------------------------------------------------------------
 */

typedef enum ModuleCallTypeTOT {
   MODULECALL_TOT_NONE = 100,
   
   MODULECALL_TOT_INTR,                  
   MODULECALL_TOT_SEMAWAIT,
   MODULECALL_TOT_SEMASIGNAL,
   MODULECALL_TOT_SEMAFORCEWAKEUP,
   MODULECALL_TOT_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_TOT_USERRETURN,
   MODULECALL_TOT_USERTIMEOUT,

   MODULECALL_TOT_GET_RECYCLED_PAGE,
   MODULECALL_TOT_RELEASE_ANON_PAGE,
   MODULECALL_TOT_SET_FASTCLOCKRATE,

   MODULECALL_TOT_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeTOT;

typedef enum ModuleCallTypeV45 {
   MODULECALL_V45_NONE = 100,
   
   MODULECALL_V45_INTR,                  
   MODULECALL_V45_SEMAWAIT,
   MODULECALL_V45_SEMASIGNAL,
   MODULECALL_V45_SEMAFORCEWAKEUP,
   MODULECALL_V45_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_V45_USERRETURN,
   MODULECALL_V45_USERTIMEOUT,

   MODULECALL_V45_GET_RECYCLED_PAGE,
   MODULECALL_V45_RELEASE_ANON_PAGE,
   MODULECALL_V45_SET_FASTCLOCKRATE,

   MODULECALL_V45_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV45;

typedef enum ModuleCallTypeV4 {
   MODULECALL_V4_NONE = 100,
   
   MODULECALL_V4_INTR,                  
   MODULECALL_V4_LOCKPAGE,                  
   MODULECALL_V4_UNLOCKPAGE,                  
   MODULECALL_V4_ISMPNLOCKED, 
   MODULECALL_V4_SEMAWAIT,
   MODULECALL_V4_SEMASIGNAL,
   MODULECALL_V4_SEMAFORCEWAKEUP,
   MODULECALL_V4_IPI,                   // hit thread with IPI

   /*
    * Return codes for user calls
    */

   MODULECALL_V4_USERRETURN,
   MODULECALL_V4_USERTIMEOUT,

   MODULECALL_V4_GET_RECYCLED_PAGE,
   MODULECALL_V4_RELEASE_ANON_PAGE,

   MODULECALL_V4_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV4;

typedef enum ModuleCallTypeV3 {
   MODULECALL_V3_NONE = 100,            //
   
   /*
    * All I/O interrupts go under MODULECALL_INTR, special-cased
    * in task switching code
    */
   MODULECALL_V3_INTR,                  
   MODULECALL_V3_LOCKPAGE,                  
   MODULECALL_V3_UNLOCKPAGE,                  

   MODULECALL_V3_ISMPNLOCKED, 
   /*
    * calls to user mode
    */
   MODULECALL_V3_USERRETURN,            // return code for USER RPC
   MODULECALL_V3_LAST                   // Number of entries. Must be the last one
} ModuleCallTypeV3;

#define MODULECALL_USER_START 300
#define MODULECALL_USER_END   399

#define MODULECALL_CROSS_PAGE_LEN    1
#define MODULECALL_CROSS_PAGE_START_V2		2
#define MODULECALL_CROSS_PAGE_START_GSX1	2
#define MODULECALL_CROSS_PAGE_START_V3		4
#define MODULECALL_CROSS_PAGE_START_V32		4
#define MODULECALL_CROSS_PAGE_START_V321	4
#define MODULECALL_CROSS_PAGE_START_GSX2	6
#define MODULECALL_CROSS_PAGE_START_GSX25	6
#define MODULECALL_CROSS_PAGE_START_V4		6
#define MODULECALL_CROSS_PAGE_START_V45		6
#define MODULECALL_CROSS_PAGE_START_TOT		6

// we don't want it in the final version
//#define USE_BLUE_SCREEN

#define MODULECALL_USERCALL_NONE     300

/*
 * Define VMX86_UCCOST in the makefiles (Local.mk,
 * typically) if you want a special build whose only purpose
 * is to measure the overhead of a user call and its
 * breakdown.
 *
 * WINDOWS NOTE: I don't know how to pass VMX86_UCCOST to
 * the driver build on Windows.  It must be defined by hand.
 *
 * ESX Note: we don't have a crosspage in which to store these
 * timestamps.  Such a feature would perhaps be nice (if we
 * ever tire of the argument that esx does so few usercalls
 * that speed doesn't matter).
 */

#if defined(VMX86_UCCOST) && !defined(VMX86_SERVER)

typedef enum UCCostStamp {
#define UC(x) UCCOST_ ## x,
#include "uccostTable.h"
   UCCOST_MAX
} UCCostStamp;

#define UCTIMESTAMP(cp, stamp) \
             do { (cp)->ucTimeStamps[UCCOST_ ## stamp] = RDTSC(); } while (0)

#else

#define UCTIMESTAMP(cp, stamp)

#endif

/*
 *----------------------------------------------------------------------
 *
 * VMCrossPage --
 *
 *      data structure shared between the monitor and the module
 *      that is used for crossing between the two.
 *      Accessible as vm->cross (kernel module) and CROSS_PAGE
 *      (monitor)
 *
 *      Exactly one page long
 *
 *----------------------------------------------------------------------
 */

/*
 * We arrived at 1008 as follows. The first 1024 bytes of the cross
 * page contains the version number (4 bytes) and then the task switch
 * code aligned at a 16 byte boundary. This alignment allows the
 * decoders to get a good start. We allocate 1024 bytes for all this
 * although the task switch code is only about 192 bytes large so
 * that the context information (data) that follows this code fragment
 * is on another 1024 byte chunk of memory. This separation makes the
 * trace cache of the P4 happy. If it any closer, the trace cache
 * gets flushed each time we write to the context info data area. 
 */
#define CONTEXT_SWITCH_CODE_SIZE_TOT 1008

#define INTERRUPT_REDIRECTION_BITMAP_SIZE 32

#define SWITCHERROR_IN_PANIC    -1
#define SWITCHERROR_NONE        0

#ifdef _WIN32
#define __attribute__(_x)     // NOP
#include <PSHPACK1.H> // 1-byte packing alignment
#endif

typedef struct Context64 {
   uint64 cr3;

   // general registers
   uint64 rax;
   uint64 rcx;
   uint64 rdx;
   uint64 rbx;
   uint64 rsi;
   uint64 rdi;
   uint64 rbp;
   uint64 rsp;

   // segment registers

   uint16 cs;
   uint16 ds;
   uint16 ss;
   uint16 es;
   uint16 fs;
   uint16 gs;

   // rip
   uint64 rip;

} __attribute__ ((packed)) Context64;

typedef struct ContextInfo {
    DTRWords    gdtr;
    DTRWords    ldtr;
    DTRWords    idtr;
    uint16 tr;
    uint8 jump[6];
    Task task;   
} __attribute__ ((packed)) ContextInfo;

typedef struct ContextInfo64 {
    DTRWords64  gdtr64;
    DTRWords64  ldtr64;
    DTRWords64  idtr64;
    uint16 tr;
    DTR64 switchGdtr64;
    FarPtr32 jump32Code;
    FarPtr32 jump64Code;
    uint16   switchDS;
    uint64   farPtr;
    Context64 context;
} __attribute__ ((packed)) ContextInfo64;

typedef struct VMCrossPageTOT {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_TOT];

   /*
    * The interrupt redirection bitmap, must immediately follow 
    * monContext.task (ASSERT fail otherwise).
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN32 crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * In x86,
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    *
    * In x86 VM on x86-64 host,
    * In oder to make the transition to legacy mode from long mode,
    * we must first turn paging off on the processor. For this we need to
    * be executing code on an identity mapped page (the crosspage). Here
    * is where we keep the phisical address of the crosspage. We will use
    * this address to create an identity mapping (a translation for which
    * PA == LA)
    *
    */
   LA crosspageLinearAddr;
   
   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeTOT moduleCallType;
   uint32 args[4];
   uint32 retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#if defined(VMX86_UCCOST) && !defined(VMX86_SERVER)
   uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   int32 switchError;

   /*
    * monUsesSysenter is TRUE iff the virtual MSR_SYSENTER_CS is
    * nonzero and the monitor uses the sysenter/sysexit MSRs for fast
    * system calls.
    */
   Bool monUsesSysenter;

   Bool _pad2[3];

   /*
    * Adjustment for machines where the hardware TSC does not run
    * constantly (laptops).  See VMK_SharedData for the ESX analog.
    * Within the monitor, PTSC_Get() = RDTSC() + tscAdjustment.  To
    * handle the case where the TSC slows or stops when the host OS
    * halts the processor using APM or ACPI, tscAdjustment is updated
    * whenever we go to userlevel.  To handle out of sync MP TSCs,
    * tscAdjustment will need to be per-PCPU and thus will need to be
    * updated whenever this VCPU migrates to a different PCPU.
    */
   VmRelativeTS tscAdjustment;

   /*
    * PTSC value and target VCPUSet for the next MonitorPoll callback,
    * as last known by the VCPU that this crosspage belongs to.  To
    * find up-to-date information, vmmon must loop through the
    * crosspages of all the VCPUs in this VM and take the one with the
    * latest time (ugh).  When the time arrives, if a target VCPU
    * thread is in the monitor, it wants to receive a hardware
    * interrupt (e.g., an IPI) as soon as possible; if it has called
    * up to userlevel to halt, it wants to wake up as soon as
    * possible.
    */
   VmAbsoluteTS monitorPollFirstTS;
   VCPUSet monitorPollFirstVCS;
   
   /*
    * Fields used by x86_64.
    *
    * Since we need to share the same VMX and VMM binary for
    * x86 and x86_64, we don't use ifdef to define x86_64
    * specific fields. If we run out of crosspage space,
    * we need to define these structures using unions.
    */
   DescriptorUnion switchGdt[4];
   ContextInfo64 hostContext64;         // replace hostContext
   Bool _pad3[2];
   uint64 hostContext64VA;              // replace hostContextVA
   uint64 hostVA64;                     
   uint64 hostSwitchCR3;

   uint64 switchHostCR4;
   uint64 hostDR[8];
   uint32 hostDRSaved;
} __attribute__ ((packed)) VMCrossPageTOT;


#ifdef WIN32
#include <POPPACK.H>
#endif


/*
 * We hardcode these constants here instead of using the sizeof() and
 * offsetof() to catch compiler inconsistencies. This file is compiled
 * once onsite to build the monitor, and then once at the customer site
 * to build the drivers.
 *
 * An alternate scheme, and preferred scheme, would be have the makefile
 * generate them, thereby relieving the developer of the error prone
 * chore of updating them each time VMCrossPage is modified. 
 *
 * Pratap 12/01.
 */
#define MODULECALL_CROSSPAGE_SIZE_TOT (0x6f4 + UCCOST_CROSSPAGE_SIZE)

#define MODULECALL_CROSSPAGE_LAOFFSET_TOT 0x54c

#define CROSSPAGE_VERSION_TOT (0x1781 + VERSION_MAGIC)

#if !defined(VMX86_SERVER) && defined(MONITOR)
#define CROSS_PAGE  ((VMCrossPage *)   \
                     (MONITOR_BASE_VADDR + VPN_2_VA(CROSS_PAGE_START)))
#define VMM_SWITCH_SHARED_DATA CROSS_PAGE
#endif

#define NULLPAGE_LINEAR_START  (MONITOR_LINEAR_START + PAGE_SIZE*NULLPAGE_START)

#define MAX_PASSTHROUGH_PORT 1024
#define NUM_PASSTHROUGH_WORDS (MAX_PASSTHROUGH_PORT / 32)

/* ---------------------------- VMware 4.5 --------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V45 1008

typedef struct VMCrossPageV45 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V45];

   /*
    * The interrupt redirection bitmap, must immediately follow 
    * monContext.task (ASSERT fail otherwise).
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN32 crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * In x86,
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    *
    * In x86 VM on x86-64 host,
    * In oder to make the transition to legacy mode from long mode,
    * we must first turn paging off on the processor. For this we need to
    * be executing code on an identity mapped page (the crosspage). Here
    * is where we keep the phisical address of the crosspage. We will use
    * this address to create an identity mapping (a translation for which
    * PA == LA)
    *
    */
   LA crosspageLinearAddr;
   
   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeV45 moduleCallType;
   uint32 args[4];
   uint32 retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA32 userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#if defined(VMX86_UCCOST) && !defined(VMX86_SERVER)
   uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   int32 switchError;

   /*
    * monUsesSysenter is TURE iff the virtual MSR_SYSENTER_CS is
    * nonzero and the monitor uses the sysenter/sysexit MSRs
    * for fast system calls.
    */
   Bool monUsesSysenter;

   Bool _pad2[3];

   /*
    * Adjustment for machines where the hardware TSC does not run
    * constantly (laptops).  See VMK_SharedData for the ESX analog.
    * Within the monitor, PTSC_Get() = RDTSC() + tscAdjustment.  To
    * handle the case where the TSC slows or stops when the host OS
    * halts the processor using APM or ACPI, tscAdjustment is updated
    * whenever we go to userlevel.  To handle out of sync MP TSCs,
    * tscAdjustment will need to be per-PCPU and thus will need to be
    * updated whenever this VCPU migrates to a different PCPU.
    */
   VmRelativeTS tscAdjustment;

   /*
    * PTSC value and target VCPUSet for the next MonitorPoll callback,
    * as last known by the VCPU that this crosspage belongs to.  To
    * find up-to-date information, vmmon must loop through the
    * crosspages of all the VCPUs in this VM and take the one with the
    * latest time (ugh).  When the time arrives, if a target VCPU
    * thread is in the monitor, it wants to receive a hardware
    * interrupt (e.g., an IPI) as soon as possible; if it has called
    * up to userlevel to halt, it wants to wake up as soon as
    * possible.
    */
   VmAbsoluteTS monitorPollFirstTS;
   VCPUSet monitorPollFirstVCS;
   
   /*
    * Fields used by x86_64.
    *
    * Since we need to share the same VMX and VMM binary for
    * x86 and x86_64, we don't use ifdef to define x86_64
    * specific fields. If we run out of crosspage space,
    * we need to define these structures using unions.
    */
   DescriptorUnion switchGdt[4];
   ContextInfo64 hostContext64;         // replace hostContext
   Bool _pad3[2];
   uint64 hostContext64VA;              // replace hostContextVA
   uint64 hostVA64;                     
   uint64 hostSwitchCR3;

   uint64 switchHostCR4;
   uint64 hostDR[8];
   uint32 hostDRSaved;
} __attribute__ ((packed)) VMCrossPageV45;

#define MODULECALL_CROSSPAGE_SIZE_V45 (0x6f4 + UCCOST_CROSSPAGE_SIZE)
#define MODULECALL_CROSSPAGE_LAOFFSET_V45 0x54c

#define CROSSPAGE_VERSION_V45 (0x1781 + VERSION_MAGIC)

/* ---------------------------- VMware 4 --------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V4 1008

typedef struct VMCrossPageV4 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V4];

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPageV4 *hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;

   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;

   MPN crossMonPageTableMPN;

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   /*
    * The monitor may requests up to two actions when
    * returning to the host.  The moduleCallType field and
    * args encode a request for some action in the driver.
    * The userCallType field (together with the RPC block)
    * encodes a user call request.  The two requests are
    * independent.  The user call is executed first, with
    * the exception of MODULECALL_INTR which has a special
    * effect.
    */
   ModuleCallTypeV4 moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   int userCallType;
   volatile int userCallRequest;	// VCPU/VMX synchronization
   Bool userCallCross;
   Bool userCallRestart;
   
   /*
    * TRUE if moduleCall was interrupted by signal. Only
    * vmmon uses this field to remember that it should
    * restart RunVM call, nobody else should look at it.
    */
   Bool moduleCallInterrupted;
   Bool _pad1[1];

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2];

#ifdef VMX86_UCCOST
    uint64 ucTimeStamps[UCCOST_MAX];
#define UCCOST_CROSSPAGE_SIZE (8 * UCCOST_MAX)
#else
#define UCCOST_CROSSPAGE_SIZE 0
#endif

   /*
    * The values in the shadow debug registers must match
    * those in the hardware debug register immediately after
    * a task switch in either direction.  They are used to
    * minimize moves to and from the debug registers.
    */
   uint32 shadowDR[8];
   uint32 switchPanic;
   Bool monUsesSysenter;

} __attribute__ ((packed)) VMCrossPageV4;

#define MODULECALL_CROSSPAGE_SIZE_V4 (0x629 + UCCOST_CROSSPAGE_SIZE)
#define MODULECALL_CROSSPAGE_LAOFFSET_V4 0x5d0

#define CROSSPAGE_VERSION_V4 (0x177d + VERSION_MAGIC)

/* ---------------------------- GSX 2.5 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_GSX25  1008

typedef struct VMCrossPageGSX25 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_GSX25];

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPageGSX25 *hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   LA monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

} __attribute__ ((packed)) VMCrossPageGSX25;

#define MODULECALL_CROSSPAGE_SIZE_GSX25     0x5fd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX25 0x5d4

#define CROSSPAGE_VERSION_GSX25 (0x1777 + VERSION_MAGIC)

/* ---------------------------- GSX 2.0 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_GSX2  1008

typedef struct VMCrossPageGSX2 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint32 _pad[3]; /* Align the code on a 16 byte boundary */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_GSX2];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   /* Kernel virtual address of the crosspage --hpreg */
   struct VMCrossPageGSX2 *hostVA;

   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   /*
    * Most significant bits (10 if the host is not in PAE mode, 11 if the host
    * is in PAE mode) of the linear address of the crosspage, which allow to
    * locate the page directory entry associated to the crosspage --hpreg
    */
   uint32 switchLinearRangeIndex;
  
   MPN crossMonPageTableMPN; 

   /* PDE that points to switch page table */
   VM_PDE  switchPDE;
   VM_PAE_PDE paeSwitchPDE; 

   LA monitorLinearAddr;      // base location of monitor in linear space
  
   /*
    * Linear address of the crosspage. It must be the same in both the kernel
    * and the monitor address spaces, because the context switch code must
    * continue to run after it flips CR3 --hpreg
    */
   LA crosspageLinearAddr;

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uintptr_t args[4];
   uintptr_t retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

   Bool monUsesSysenter;

} __attribute__ ((packed)) VMCrossPageGSX2;

#define MODULECALL_CROSSPAGE_SIZE_GSX2     0x5dd
#define MODULECALL_CROSSPAGE_LAOFFSET_GSX2 0x5b4

#define CROSSPAGE_VERSION_GSX2 (0x1777 + VERSION_MAGIC)

/* ---------------------------- WS 3.2.1 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V321	1008

typedef struct VMCrossPageV321 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
   uint32 _pad[3];

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V321];

   /*
    * iobitMap must immediately follow the interrupt redirection
    * bitmap, which must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint8       interruptRedirectionBitMap[INTERRUPT_REDIRECTION_BITMAP_SIZE];
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPageV321 * hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
#endif
} __attribute__ ((packed)) VMCrossPageV321;

#define MODULECALL_CROSSPAGE_SIZE_V321     0x5f0
#define MODULECALL_CROSSPAGE_LAOFFSET_V321 0x5c8

#define CROSSPAGE_VERSION_V321 (0x1776 + VERSION_MAGIC)

/* ---------------------------- WS 3.2 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V32	1008

typedef struct VMCrossPageV32 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
   uint32 _pad[3];

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V32];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPageV32 * hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
#endif
} __attribute__ ((packed)) VMCrossPageV32;

#define MODULECALL_CROSSPAGE_SIZE_V32     0x5d0
#define MODULECALL_CROSSPAGE_LAOFFSET_V32 0x5a8

#define CROSSPAGE_VERSION_V32 (0x1776 + VERSION_MAGIC)

/* ---------------------------- WS 3.0 ---------------------------- */

#define CONTEXT_SWITCH_CODE_SIZE_V3	192

typedef struct VMCrossPageV3 {
   /*
    * Version checking. Should remain at offset 0
    */ 
   uint32 version; 
  

   /*
    * Changes in contextSwitchCode location need to
    * be reflected in addrlayout.h:COMINGBACK_OFFSET
    * (ASSERT fail otherwise)
    */
   uint8 contextSwitchCode[CONTEXT_SWITCH_CODE_SIZE_V3];

   /*
    * iobitMap must immediately follow monContext.task
    * (ASSERT fail otherwise)
    */
   ContextInfo hostContext;
   ContextInfo monContext;
   uint32      iobitMap[NUM_PASSTHROUGH_WORDS];

   struct VMCrossPageV3 * hostVA;
   VM_PTE hostSharedPTE;

   uint32 hostContextVA;


   uint32 switchPDOffset;     // index into the page directory of switchPDE
   VM_PDE switchPDE;          // PDE that points to switch page table


   LA monitorLinearAddr;      // base location of monitor in linear space
   LA crosspageLinearAddr;    // location of crosspage page in both linear spaces

   //
   // ModuleCall parameters
   //
   ModuleCallTypeV3 moduleCallType;
   uint32 args[4];
   uint32 retval;

   VA userSharedAreaAddr;

   // host irq relocation values
   int irqRelocateOffset[2]; 

#ifdef __cplusplus
   static const int monUsesSysenter = 0;
#endif
} __attribute__ ((packed)) VMCrossPageV3;

#define MODULECALL_CROSSPAGE_SIZE_V3     0x294
#define MODULECALL_CROSSPAGE_LAOFFSET_V3 0x26c


#define MX_WAITTIMEOUT       100                             // unit: 1ms

#define MX_WAITINTERRUPTED     3
#define MX_WAITTIMEDOUT        2
#define MX_WAITNORMAL          1  // Must equal one; see linux module code.
#define MX_WAITERROR           0  // Use MX_ISWAITERROR() to test for error.

// Any zero or negative value denotes error.
#define MX_ISWAITERROR(e)      ((e) <= MX_WAITERROR)

#endif
